import cv2
import numpy as np
import os
from shutil import copy2
from plot import HsvCone
import dash
from dash import dcc, html

# converts sequence of pixels in bgr (1xMany picture)
# to average color in rgb and hsv (sum up all pixels by channel and divide by amount of pixels)
def seq2mean(seq):
    bgr_mean = np.rint(np.mean(seq, axis=0)).astype('uint8')
    hsv_mean = cv2.cvtColor(np.array([[bgr_mean]]), cv2.COLOR_BGR2HSV)
    return bgr_mean[::-1], hsv_mean.flatten()

# converts hue to its absolute color in rgb
def hue2rgb_abs(hue):
    return cv2.cvtColor(np.array([[[hue, 255, 255]]]).astype('uint8'), cv2.COLOR_HSV2RGB).flatten()

# lots of things are going on here, but in short,
# func accepts relative path to image, calculate its
# average color in rgb and hsv, than delete all gray (black and white too) pixels
# and calculate average, prevailing hue and its percentage
# if whole image is gray, image doesn't participate in sorting and plotting
def get_image_info(image_path):
    bgr_sequence = cv2.imdecode(np.fromfile(image_path, dtype=np.uint8), cv2.IMREAD_UNCHANGED)[:, :, 0:3].reshape(-1, 3)
    rgb_average_color, hsv_average_color = seq2mean(bgr_sequence)

    gray_indices = np.all(bgr_sequence == bgr_sequence[:, 0][:, np.newaxis], axis=1)
    if np.all(gray_indices):
        return -1
    nongray_bgr_sequence = bgr_sequence[~gray_indices]
    nongray_hsv_sequence = cv2.cvtColor(np.array([nongray_bgr_sequence]),
                                        cv2.COLOR_BGR2HSV)[0]
    avg_hue_hsv = seq2mean(nongray_bgr_sequence)[1][0]
    hues_count = np.bincount(nongray_hsv_sequence[:, 0], minlength=180)
    # for 2d plot
    hues = np.column_stack((np.arange(180), hues_count))
    prevailing_hue_count_index = np.argmax(hues_count)
    # for sorting
    prevailing_hue = hues[prevailing_hue_count_index, 0]
    prevailing_hue_count = hues[prevailing_hue_count_index, 1]
    # additional metrics for sorting
    prevailing_hue_percentage = np.round(prevailing_hue_count / nongray_hsv_sequence.shape[0], 4)
    # for 3d scatter points
    plot_data = dict(hsv=np.array([hsv_average_color,
                                   [avg_hue_hsv, 255, 255],
                                   [prevailing_hue, 255, 255]]),
                     rgb=np.array([rgb_average_color,
                                   hue2rgb_abs(avg_hue_hsv),
                                   hue2rgb_abs(prevailing_hue)]).astype('uint8'))
    
    return plot_data, hues, prevailing_hue_percentage, prevailing_hue


# hard-coded extensions of image files
IMAGE_EXTENSIONS = ('.bmp', '.dib', '.rle',
                    'jpg', '.jpeg', '.JPG', '.JPEG', '.jpe', '.JPE',
                    '.png',
                    '.tif', '.tiff',
                    '.webp')
sort_list, subdirs, images = [], [], []

plot = HsvCone()
for item in os.listdir():
    if item.endswith(IMAGE_EXTENSIONS):
        if (f_return := get_image_info(item)) == -1:
            print(f"Warning: file '{item}' is grayscale,",
                  "so won't be sorted and displayed on plot.")
        else:
            plot_data, hues, prevailing_hue_percentage, prevailing_hue = f_return
            # add 3d scatter points and 2d plot to general plot
            plot.add_image_data(plot_data, hues, item)
            # write images file name
            images.append(item)
            # sort_list is variable for values for sorting
            # prevailing_hue_percentage is value between 0 and 1 and hue step is 1
            # so the more pixels with prevailing hue are in image relatively
            # to overall amount of pixels, the more prior this image is in sorting 
            sort_list.append(prevailing_hue - prevailing_hue_percentage)
    # gather subdirs to not overwrite existing one
    elif os.path.isdir(item):
        subdirs.append(item)
#finally sort images
sorted_images = np.array(images)[np.argsort(sort_list)]

#checking for existing subdirs
sorted_images_dir, counter = 'sorted images', 1
while sorted_images_dir in subdirs:
    sorted_images_dir = f'sorted images ({counter})'
    counter += 1
os.mkdir(sorted_images_dir)
# copy images with new names
for i, image in enumerate(sorted_images):
    copy2(image, f"{sorted_images_dir}/{i+1}.{image.split('.')[-1]}")

# dash app initialization
app = dash.Dash(__name__)
app.layout = html.Div(dcc.Graph(id='plot', style={'width': '99vw', 'height': '98vh'}, figure=plot.fig))
@app.callback(
    dash.dependencies.Output('plot', 'figure'),
    dash.dependencies.Output('plot', 'clickData'),
    [dash.dependencies.Input('plot', 'clickData')],
    [dash.dependencies.State('plot', 'figure')])
# 'click' event handler, shows and hides 2d plot if clicked on 3d scatter point
def show_hide_2d_plot(clickData, figure):
    if clickData is not None:
        trace_index = clickData['points'][0]['curveNumber']
        if not (trace_index < 4 or trace_index % 2):
            new_data = list(figure['data'])
            match new_data[trace_index+1]['visible']:
                case True:
                    new_data[trace_index+1]['visible'] = 'legendonly'
                case 'legendonly':
                    new_data[trace_index+1]['visible'] = True
    return figure, None

# starting local server for plot
app.run_server(debug=True, use_reloader=False)
