import plotly.graph_objects as go
from plotly.express.colors import cyclical
import numpy as np
from plotly.subplots import make_subplots


class HsvCone:
    # build a cone of HSV color space
    def __init__(self) -> None:
        # cone parametrical equations:
        r, theta = np.mgrid[0:255:256j, 0:2*np.pi:360j]
        x = r*np.cos(theta)
        y = r*np.sin(theta)
        z = r
        z_255 = 255*np.ones(x.shape)

        # plotly has limitations about coloring surfaces,
        # so, to implement 'value' and 'saturation' gradients,
        # cone and circle are doubled and colored to hsv colors,
        # and then to white-transparent and black-transparent gradient
        radial_colors = np.arctan2(y, x)
        black = [[0, 'black'], [1, 'black']]
        white = [[0, 'white'], [1, 'white']]
        # fading of gradient is not linear due to rendering peculiarities
        three_halves_fading = [[i, (1-i)**1.5] for i in np.linspace(0, 1, num=10)]
        quintic_fading = [[i, (1-i)**5] for i in np.linspace(0, 1)]

        fig = make_subplots(rows=1, cols=2, specs=[[{'type': 'surface'}, {'type': 'scatter'}]],
                            subplot_titles=['HSV Cone', 'Hue distribution'])
        fig.add_trace(go.Surface(x=x, y=y, z=z, hoverinfo='none', showscale=False,
                                 opacity=.7, surfacecolor=radial_colors, colorscale=cyclical.HSV))
        fig.add_trace(go.Surface(x=x, y=y, z=z, hoverinfo='none', showscale=False,
                                 opacityscale=three_halves_fading, colorscale=black))
        fig.add_trace(go.Surface(x=x, y=y, z=z_255, hoverinfo='none', showscale=False,
                                 opacity=.3, surfacecolor=radial_colors, colorscale=cyclical.HSV))
        fig.add_trace(go.Surface(x=x, y=y, z=z_255, hoverinfo='none', showscale=False,
                                 opacityscale=quintic_fading, surfacecolor=x**2+y**2, colorscale=white))
        fig.update_layout(margin=dict(l=50, r=50, t=30, b=30), xaxis_title="Hue", yaxis_title="Pixels count",
                          scene = dict(xaxis_title='', yaxis_title='', zaxis_title='Value'))

        self._fig = fig

    # adding 3 3dscatter points to the cone, and 2d plot of hue distribution
    # points are: average color, average hue, prevailing hue
    def add_image_data(self, data_3d: dict, data_2d: np.ndarray, name: str) -> None:
        hsv = data_3d['hsv']
        rgb = data_3d['rgb']
        # calculate coordinates of 3d scatter points from HSV values
        x = -hsv[:, 1]*hsv[:, 2]*np.cos(np.pi*hsv[:, 0]/90)/255
        y = -hsv[:, 1]*hsv[:, 2]*np.sin(np.pi*hsv[:, 0]/90)/255
        z = hsv[:, 2]
        # plotly has a bug of incorrect hsv color interpretation (my Github issue: https://github.com/plotly/plotly.py/issues/4382)
        # so rgb colors used instead to color points
        colors = [f'rgb({i[0]}, {i[1]}, {i[2]})' for i in rgb]
        # custom hovertemplate for 3d scatter points to show hue of current point
        # and average color of image in HSV format
        hovertemplate =''.join(('<b>H: %{customdata[0]}</b><br>',
                                '<i>avg:</i><br>',
                                'H: %{customdata[1]}<br>',
                                'S: %{customdata[2]}<br>'
                                'V: %{customdata[3]}'))
        customdata = tuple((hsv[i, 0], *hsv[0, :]) for i in range(3))

        dots_3d = go.Scatter3d(x=x, y=y, z=z, customdata=customdata, hovertemplate=hovertemplate,
                               hoverlabel=dict(bgcolor=colors[0]), name=name,
                               marker=dict(size=(15,15,25), color=colors, opacity=.9))
        dots_2d = go.Scatter(x=data_2d[:, 0], y=data_2d[:, 1], name=name, line_color=colors[0],
                             visible='legendonly', mode='lines+markers',
                             marker=dict(cmin=0, cmax=179, color=data_2d[:, 0],
                                         colorscale=cyclical.HSV))
        
        self._fig.add_traces(data=[dots_3d, dots_2d])

    # simple figure getter
    @property
    def fig(self) -> go.Figure:
        return self._fig
